-module(spoke).
-compile([no_auto_import, nowarn_unused_vars, nowarn_unused_function, nowarn_nomatch]).

-export([connect_with_id/2, keep_alive_seconds/2, server_timeout_ms/2, using_auth/3, updates/1, connect/2, connect_with_will/3, disconnect/1, publish/2, pending_publishes/1, wait_for_publishes_to_finish/2, subscribe/2, unsubscribe/2, unlink_and_crash/1, start_session_with_ms_keep_alive/6, start_session/1, restore_session/2]).
-export_type([client/0, auth_details/0, connect_options/0, update/0, connection_state/0, publish_data/0, operation_error/0, subscription/0, qo_s/0, connect_error/0, subscribe_request/0, config/0, state/0, message/0, pending_subscription/0]).

-if(?OTP_RELEASE >= 27).
-define(MODULEDOC(Str), -moduledoc(Str)).
-define(DOC(Str), -doc(Str)).
-else.
-define(MODULEDOC(Str), -compile([])).
-define(DOC(Str), -compile([])).
-endif.

-opaque client() :: {client,
        gleam@erlang@process:subject(message()),
        gleam@erlang@process:subject(update()),
        config()}.

-type auth_details() :: {auth_details,
        binary(),
        gleam@option:option(bitstring())}.

-type connect_options() :: {connect_options,
        fun(() -> {ok,
                {fun((gleam@bytes_tree:bytes_tree()) -> {ok, nil} |
                        {error, binary()}),
                    fun(() -> gleam@erlang@process:selector({ok, bitstring()} |
                        {error, binary()})),
                    fun(() -> nil)}} |
            {error, binary()}),
        binary(),
        gleam@option:option(auth_details()),
        integer(),
        integer()}.

-type update() :: {received_message, binary(), bitstring(), boolean()} |
    {connection_state_changed, connection_state()}.

-type connection_state() :: {connect_failed, binary()} |
    {connect_rejected, connect_error()} |
    {connect_accepted, boolean()} |
    disconnected |
    {disconnected_unexpectedly, binary()}.

-type publish_data() :: {publish_data, binary(), bitstring(), qo_s(), boolean()}.

-type operation_error() :: not_connected |
    operation_timed_out |
    protocol_violation |
    killed_during_operation.

-type subscription() :: {successful_subscription, binary(), qo_s()} |
    {failed_subscription, binary()}.

-type qo_s() :: at_most_once | at_least_once | exactly_once.

-type connect_error() :: unacceptable_protocol_version |
    identifier_refused |
    server_unavailable |
    bad_username_or_password |
    not_authorized.

-type subscribe_request() :: {subscribe_request, binary(), qo_s()}.

-type config() :: {config,
        binary(),
        gleam@option:option(auth_details()),
        integer(),
        integer(),
        fun(() -> {ok, spoke@internal@transport:byte_channel()} |
            {error, binary()})}.

-type state() :: {state,
        gleam@erlang@process:subject(message()),
        config(),
        gleam@erlang@process:subject(update()),
        gleam@option:option(spoke@internal@connection:connection()),
        boolean(),
        spoke@internal@session:session(),
        gleam@dict:dict(integer(), pending_subscription()),
        gleam@dict:dict(integer(), gleam@erlang@process:subject({ok, nil} |
            {error, operation_error()})),
        list(gleam@erlang@process:subject(nil))}.

-type message() :: crash |
    {connect, boolean(), gleam@option:option(publish_data())} |
    {publish, publish_data()} |
    {subscribe,
        list(subscribe_request()),
        gleam@erlang@process:subject({ok, list(subscription())} |
            {error, operation_error()})} |
    {unsubscribe,
        list(binary()),
        gleam@erlang@process:subject({ok, nil} | {error, operation_error()})} |
    {process_received, spoke@internal@packet@client@incoming:packet()} |
    {connection_dropped, spoke@internal@connection:disconnect()} |
    {disconnect, gleam@erlang@process:subject(binary())} |
    {drop_connection_and_notify_client, binary()} |
    {get_pending_publishes, gleam@erlang@process:subject(integer())} |
    {wait_for_publishes_to_finish, gleam@erlang@process:subject(nil)}.

-type pending_subscription() :: {pending_subscription,
        list(subscribe_request()),
        gleam@erlang@process:subject({ok, list(subscription())} |
            {error, operation_error()})}.

-file("src/spoke.gleam", 158).
?DOC(
    " Constructs connect options from transport options, the given client id,\n"
    " and default settings for the rest of the options.\n"
).
-spec connect_with_id(
    fun(() -> {ok,
            {fun((gleam@bytes_tree:bytes_tree()) -> {ok, nil} |
                    {error, binary()}),
                fun(() -> gleam@erlang@process:selector({ok, bitstring()} |
                    {error, binary()})),
                fun(() -> nil)}} |
        {error, binary()}),
    binary()
) -> connect_options().
connect_with_id(Connector, Client_id) ->
    {connect_options, Connector, Client_id, none, 15, 5000}.

-file("src/spoke.gleam", 172).
?DOC(" Builder function for specifying the keep-alive time in the connect options.\n").
-spec keep_alive_seconds(connect_options(), integer()) -> connect_options().
keep_alive_seconds(Options, Keep_alive_seconds) ->
    _record = Options,
    {connect_options,
        erlang:element(2, _record),
        erlang:element(3, _record),
        erlang:element(4, _record),
        Keep_alive_seconds,
        erlang:element(6, _record)}.

-file("src/spoke.gleam", 180).
?DOC(" Builder function for specifying the server operation timeout in the connect options.\n").
-spec server_timeout_ms(connect_options(), integer()) -> connect_options().
server_timeout_ms(Options, Server_timeout_ms) ->
    _record = Options,
    {connect_options,
        erlang:element(2, _record),
        erlang:element(3, _record),
        erlang:element(4, _record),
        erlang:element(5, _record),
        Server_timeout_ms}.

-file("src/spoke.gleam", 188).
?DOC(" Builder function for specifying the authentication details to be used when connecting.\n").
-spec using_auth(connect_options(), binary(), gleam@option:option(bitstring())) -> connect_options().
using_auth(Options, Username, Password) ->
    _record = Options,
    {connect_options,
        erlang:element(2, _record),
        erlang:element(3, _record),
        {some, {auth_details, Username, Password}},
        erlang:element(5, _record),
        erlang:element(6, _record)}.

-file("src/spoke.gleam", 238).
?DOC(
    " Returns a `Subject` for receiving client updates\n"
    " (received messages and connection state changes).\n"
    " NOTE: This function is provided for convenience\n"
    " as receiving updates only works from the process that created the client!\n"
    " Future versions of spoke might support receiving updates in multiple processes.\n"
).
-spec updates(client()) -> gleam@erlang@process:subject(update()).
updates(Client) ->
    erlang:element(3, Client).

-file("src/spoke.gleam", 248).
?DOC(
    " Starts connecting to the MQTT server.\n"
    " The connection state will be published as an update.\n"
    " If a connection is already established or being established,\n"
    " this will be a no-op.\n"
    " Note that switching between `clean_session` values\n"
    " while already connecting is currently not well handled.\n"
).
-spec connect(client(), boolean()) -> nil.
connect(Client, Clean_session) ->
    gleam@erlang@process:send(
        erlang:element(2, Client),
        {connect, Clean_session, none}
    ).

-file("src/spoke.gleam", 253).
?DOC(" Same as `connect`, but specifies a will message for this connection.\n").
-spec connect_with_will(client(), boolean(), publish_data()) -> nil.
connect_with_will(Client, Clean_session, Will) ->
    gleam@erlang@process:send(
        erlang:element(2, Client),
        {connect, Clean_session, {some, Will}}
    ).

-file("src/spoke.gleam", 266).
?DOC(
    " Disconnects from the MQTT server.\n"
    " The connection state change will also be published as an update.\n"
    " If a connection is not established or being established,\n"
    " this will be a no-op.\n"
    " Returns the serialized session state to be potentially restored later.\n"
).
-spec disconnect(client()) -> binary().
disconnect(Client) ->
    gleam@erlang@process:call(
        erlang:element(2, Client),
        fun(_capture) -> {disconnect, _capture} end,
        erlang:element(5, erlang:element(4, Client))
    ).

-file("src/spoke.gleam", 275).
?DOC(
    " Connects to the MQTT server.\n"
    " Will disconnect if the connect times out,\n"
    " and send the `Disconnect` update.\n"
    " Note that in case of a timeout,\n"
    " the message might still have been already published.\n"
).
-spec publish(client(), publish_data()) -> nil.
publish(Client, Data) ->
    gleam@erlang@process:send(erlang:element(2, Client), {publish, Data}).

-file("src/spoke.gleam", 301).
?DOC(
    " Returns the number of QoS > 0 publishes that haven't yet been completely published.\n"
    " Also see `wait_for_publishes_to_finish`.\n"
).
-spec pending_publishes(client()) -> integer().
pending_publishes(Client) ->
    gleam@erlang@process:call(
        erlang:element(2, Client),
        fun(_capture) -> {get_pending_publishes, _capture} end,
        erlang:element(5, erlang:element(4, Client))
    ).

-file("src/spoke.gleam", 312).
?DOC(
    " Wait for all pending QoS > 0 publishes to complete.\n"
    " Returns an error if the operation times out,\n"
    " or client is killed while waiting.\n"
).
-spec wait_for_publishes_to_finish(client(), integer()) -> {ok, nil} |
    {error, nil}.
wait_for_publishes_to_finish(Client, Timeout) ->
    _pipe = gleam@erlang@process:try_call(
        erlang:element(2, Client),
        fun(_capture) -> {wait_for_publishes_to_finish, _capture} end,
        Timeout
    ),
    gleam@result:replace_error(_pipe, nil).

-file("src/spoke.gleam", 320).
-spec call_or_disconnect(
    client(),
    fun((gleam@erlang@process:subject({ok, QPJ} | {error, operation_error()})) -> message())
) -> {ok, QPJ} | {error, operation_error()}.
call_or_disconnect(Client, Make_request) ->
    _pipe = gleam@erlang@process:try_call(
        erlang:element(2, Client),
        Make_request,
        erlang:element(5, erlang:element(4, Client))
    ),
    _pipe@1 = gleam@result:map_error(_pipe, fun(E) -> case E of
                call_timeout ->
                    gleam@erlang@process:send(
                        erlang:element(2, Client),
                        {drop_connection_and_notify_client,
                            <<"Operation timed out"/utf8>>}
                    ),
                    operation_timed_out;

                {callee_down, _} ->
                    killed_during_operation
            end end),
    gleam@result:flatten(_pipe@1).

-file("src/spoke.gleam", 282).
?DOC(
    " Subscribes to the given topics.\n"
    " Will block until we get a response from the server,\n"
    " returning the result of the operation.\n"
).
-spec subscribe(client(), list(subscribe_request())) -> {ok,
        list(subscription())} |
    {error, operation_error()}.
subscribe(Client, Topics) ->
    call_or_disconnect(
        Client,
        fun(_capture) -> {subscribe, Topics, _capture} end
    ).

-file("src/spoke.gleam", 292).
?DOC(
    " Unsubscribes from the given topics.\n"
    " Will block until we get a response from the server,\n"
    " returning the result of the operation.\n"
).
-spec unsubscribe(client(), list(binary())) -> {ok, nil} |
    {error, operation_error()}.
unsubscribe(Client, Topics) ->
    call_or_disconnect(
        Client,
        fun(_capture) -> {unsubscribe, Topics, _capture} end
    ).

-file("src/spoke.gleam", 369).
?DOC(false).
-spec unlink_and_crash(client()) -> nil.
unlink_and_crash(Client) ->
    gleam@erlang@process:unlink(
        gleam@erlang@process:subject_owner(erlang:element(2, Client))
    ),
    gleam@erlang@process:send(erlang:element(2, Client), crash).

-file("src/spoke.gleam", 425).
-spec init(
    config(),
    spoke@internal@session:session(),
    gleam@erlang@process:subject(update())
) -> gleam@otp@actor:init_result(state(), message()).
init(Config, Session, Updates) ->
    Self = gleam@erlang@process:new_subject(),
    State = {state,
        Self,
        Config,
        Updates,
        none,
        false,
        Session,
        maps:new(),
        maps:new(),
        []},
    Selector = begin
        _pipe = gleam_erlang_ffi:new_selector(),
        gleam@erlang@process:selecting(
            _pipe,
            Self,
            fun gleam@function:identity/1
        )
    end,
    {ready, State, Selector}.

-file("src/spoke.gleam", 470).
-spec get_pending_publishes(state(), gleam@erlang@process:subject(integer())) -> gleam@otp@actor:next(message(), state()).
get_pending_publishes(State, Reply_to) ->
    Pending_pubs = spoke@internal@session:pending_publishes(
        erlang:element(7, State)
    ),
    gleam@erlang@process:send(Reply_to, Pending_pubs),
    gleam@otp@actor:continue(State).

-file("src/spoke.gleam", 479).
-spec handle_wait_for_publishes_to_finish(
    state(),
    gleam@erlang@process:subject(nil)
) -> gleam@otp@actor:next(any(), state()).
handle_wait_for_publishes_to_finish(State, Reply_to) ->
    Publish_completion_listeners = [Reply_to | erlang:element(10, State)],
    gleam@otp@actor:continue(
        begin
            _record = State,
            {state,
                erlang:element(2, _record),
                erlang:element(3, _record),
                erlang:element(4, _record),
                erlang:element(5, _record),
                erlang:element(6, _record),
                erlang:element(7, _record),
                erlang:element(8, _record),
                erlang:element(9, _record),
                Publish_completion_listeners}
        end
    ).

-file("src/spoke.gleam", 487).
-spec handle_connection_drop(state(), spoke@internal@connection:disconnect()) -> gleam@otp@actor:next(message(), state()).
handle_connection_drop(State, Reason) ->
    Connecting = not erlang:element(6, State),
    Change = case Reason of
        {abnormal_disconnect, Reason@1} when Connecting ->
            {some, {connect_failed, Reason@1}};

        graceful_disconnect when Connecting ->
            none;

        {abnormal_disconnect, Reason@2} ->
            {some, {disconnected_unexpectedly, Reason@2}};

        graceful_disconnect ->
            {some, disconnected}
    end,
    case Change of
        {some, Change@1} ->
            gleam@erlang@process:send(
                erlang:element(4, State),
                {connection_state_changed, Change@1}
            );

        none ->
            nil
    end,
    gleam@otp@actor:continue(
        begin
            _record = State,
            {state,
                erlang:element(2, _record),
                erlang:element(3, _record),
                erlang:element(4, _record),
                none,
                false,
                erlang:element(7, _record),
                erlang:element(8, _record),
                erlang:element(9, _record),
                erlang:element(10, _record)}
        end
    ).

-file("src/spoke.gleam", 769).
-spec handle_incoming_publish(state(), spoke@internal@packet:publish_data()) -> gleam@otp@actor:next(message(), state()).
handle_incoming_publish(State, Data) ->
    {Msg@4, Session@1, Packet} = case Data of
        {publish_data_qo_s0, Msg} ->
            {{some, Msg}, erlang:element(7, State), none};

        {publish_data_qo_s1, Msg@1, _, Id} ->
            {{some, Msg@1}, erlang:element(7, State), {some, {pub_ack, Id}}};

        {publish_data_qo_s2, Msg@2, _, Id@1} ->
            {Session, Publish_result} = spoke@internal@session:start_qos2_receive(
                erlang:element(7, State),
                Id@1
            ),
            Msg@3 = case Publish_result of
                true ->
                    {some, Msg@2};

                false ->
                    none
            end,
            {Msg@3, Session, {some, {pub_rec, Id@1}}}
    end,
    case {Packet, erlang:element(5, State)} of
        {{some, Packet@1}, {some, Connection}} ->
            spoke@internal@connection:send(Connection, Packet@1);

        {_, _} ->
            nil
    end,
    case Msg@4 of
        {some, Msg@5} ->
            Update = {received_message,
                erlang:element(2, Msg@5),
                erlang:element(3, Msg@5),
                erlang:element(4, Msg@5)},
            gleam@erlang@process:send(erlang:element(4, State), Update);

        none ->
            nil
    end,
    gleam@otp@actor:continue(
        begin
            _record = State,
            {state,
                erlang:element(2, _record),
                erlang:element(3, _record),
                erlang:element(4, _record),
                erlang:element(5, _record),
                erlang:element(6, _record),
                Session@1,
                erlang:element(8, _record),
                erlang:element(9, _record),
                erlang:element(10, _record)}
        end
    ).

-file("src/spoke.gleam", 811).
-spec handle_disconnect(state(), gleam@erlang@process:subject(binary())) -> gleam@otp@actor:next(message(), state()).
handle_disconnect(State, Reply_to) ->
    Result = case erlang:element(5, State) of
        {some, Connection} ->
            spoke@internal@connection:send(Connection, disconnect),
            spoke@internal@connection:shutdown(Connection),
            gleam@otp@actor:continue(
                begin
                    _record = State,
                    {state,
                        erlang:element(2, _record),
                        erlang:element(3, _record),
                        erlang:element(4, _record),
                        none,
                        erlang:element(6, _record),
                        erlang:element(7, _record),
                        erlang:element(8, _record),
                        erlang:element(9, _record),
                        erlang:element(10, _record)}
                end
            );

        none ->
            gleam@otp@actor:continue(State)
    end,
    gleam@erlang@process:send(
        Reply_to,
        spoke@internal@session:to_json(erlang:element(7, State))
    ),
    Result.

-file("src/spoke.gleam", 828).
-spec notify_if_publishes_finished(state()) -> state().
notify_if_publishes_finished(State) ->
    case spoke@internal@session:pending_publishes(erlang:element(7, State)) of
        0 ->
            begin
                gleam@list:each(
                    erlang:element(10, State),
                    fun(Listener) ->
                        gleam@erlang@process:send(Listener, nil)
                    end
                )
            end,
            _record = State,
            {state,
                erlang:element(2, _record),
                erlang:element(3, _record),
                erlang:element(4, _record),
                erlang:element(5, _record),
                erlang:element(6, _record),
                erlang:element(7, _record),
                erlang:element(8, _record),
                erlang:element(9, _record),
                []};

        _ ->
            State
    end.

-file("src/spoke.gleam", 719).
-spec handle_pubcomp(state(), integer()) -> gleam@otp@actor:next(message(), state()).
handle_pubcomp(State, Id) ->
    Session = spoke@internal@session:handle_pubcomp(
        erlang:element(7, State),
        Id
    ),
    State@1 = begin
        _pipe = begin
            _record = State,
            {state,
                erlang:element(2, _record),
                erlang:element(3, _record),
                erlang:element(4, _record),
                erlang:element(5, _record),
                erlang:element(6, _record),
                Session,
                erlang:element(8, _record),
                erlang:element(9, _record),
                erlang:element(10, _record)}
        end,
        notify_if_publishes_finished(_pipe)
    end,
    gleam@otp@actor:continue(State@1).

-file("src/spoke.gleam", 841).
-spec guard_connected(
    state(),
    fun(() -> gleam@otp@actor:next(message(), state())),
    fun((spoke@internal@connection:connection()) -> gleam@otp@actor:next(message(), state()))
) -> gleam@otp@actor:next(message(), state()).
guard_connected(State, If_not_connected, If_connected) ->
    case erlang:element(5, State) of
        none ->
            If_not_connected();

        {some, Connection} ->
            If_connected(Connection)
    end.

-file("src/spoke.gleam", 581).
-spec handle_outgoing_publish(state(), publish_data()) -> gleam@otp@actor:next(message(), state()).
handle_outgoing_publish(State, Data) ->
    Message = {message_data,
        erlang:element(2, Data),
        erlang:element(3, Data),
        erlang:element(5, Data)},
    {Session, Packet} = case erlang:element(4, Data) of
        at_most_once ->
            {erlang:element(7, State), {publish, {publish_data_qo_s0, Message}}};

        at_least_once ->
            spoke@internal@session:start_qos1_publish(
                erlang:element(7, State),
                Message
            );

        exactly_once ->
            spoke@internal@session:start_qos2_publish(
                erlang:element(7, State),
                Message
            )
    end,
    State@1 = begin
        _record = State,
        {state,
            erlang:element(2, _record),
            erlang:element(3, _record),
            erlang:element(4, _record),
            erlang:element(5, _record),
            erlang:element(6, _record),
            Session,
            erlang:element(8, _record),
            erlang:element(9, _record),
            erlang:element(10, _record)}
    end,
    guard_connected(
        State@1,
        fun() -> gleam@otp@actor:continue(State@1) end,
        fun(Connection) ->
            spoke@internal@connection:send(Connection, Packet),
            gleam@otp@actor:continue(State@1)
        end
    ).

-file("src/spoke.gleam", 701).
-spec handle_pubrel(state(), integer()) -> gleam@otp@actor:next(message(), state()).
handle_pubrel(State, Id) ->
    guard_connected(
        State,
        fun() -> gleam@otp@actor:continue(State) end,
        fun(Connection) ->
            spoke@internal@connection:send(Connection, {pub_comp, Id}),
            Session = spoke@internal@session:handle_pubrel(
                erlang:element(7, State),
                Id
            ),
            gleam@otp@actor:continue(
                begin
                    _record = State,
                    {state,
                        erlang:element(2, _record),
                        erlang:element(3, _record),
                        erlang:element(4, _record),
                        erlang:element(5, _record),
                        erlang:element(6, _record),
                        Session,
                        erlang:element(8, _record),
                        erlang:element(9, _record),
                        erlang:element(10, _record)}
                end
            )
        end
    ).

-file("src/spoke.gleam", 711).
-spec handle_pubrec(state(), integer()) -> gleam@otp@actor:next(message(), state()).
handle_pubrec(State, Id) ->
    guard_connected(
        State,
        fun() -> gleam@otp@actor:continue(State) end,
        fun(Connection) ->
            Session = spoke@internal@session:handle_pubrec(
                erlang:element(7, State),
                Id
            ),
            spoke@internal@connection:send(Connection, {pub_rel, Id}),
            gleam@otp@actor:continue(
                begin
                    _record = State,
                    {state,
                        erlang:element(2, _record),
                        erlang:element(3, _record),
                        erlang:element(4, _record),
                        erlang:element(5, _record),
                        erlang:element(6, _record),
                        Session,
                        erlang:element(8, _record),
                        erlang:element(9, _record),
                        erlang:element(10, _record)}
                end
            )
        end
    ).

-file("src/spoke.gleam", 751).
-spec handle_unsubscribe(
    state(),
    list(binary()),
    gleam@erlang@process:subject({ok, nil} | {error, operation_error()})
) -> gleam@otp@actor:next(message(), state()).
handle_unsubscribe(State, Topics, Reply_to) ->
    guard_connected(
        State,
        fun() ->
            gleam@erlang@process:send(Reply_to, {error, not_connected}),
            gleam@otp@actor:continue(State)
        end,
        fun(Connection) ->
            {Session, Id} = spoke@internal@session:reserve_packet_id(
                erlang:element(7, State)
            ),
            Pending_unsubs = begin
                _pipe = erlang:element(9, State),
                gleam@dict:insert(_pipe, Id, Reply_to)
            end,
            spoke@internal@connection:send(
                Connection,
                {unsubscribe, Id, Topics}
            ),
            gleam@otp@actor:continue(
                begin
                    _record = State,
                    {state,
                        erlang:element(2, _record),
                        erlang:element(3, _record),
                        erlang:element(4, _record),
                        erlang:element(5, _record),
                        erlang:element(6, _record),
                        Session,
                        erlang:element(8, _record),
                        Pending_unsubs,
                        erlang:element(10, _record)}
                end
            )
        end
    ).

-file("src/spoke.gleam", 852).
-spec drop_connection_and_notify(state(), gleam@option:option(binary())) -> gleam@otp@actor:next(message(), state()).
drop_connection_and_notify(State, Error) ->
    case erlang:element(5, State) of
        {some, Connection} ->
            spoke@internal@connection:shutdown(Connection),
            Update = case Error of
                none ->
                    disconnected;

                {some, Reason} ->
                    {disconnected_unexpectedly, Reason}
            end,
            gleam@erlang@process:send(
                erlang:element(4, State),
                {connection_state_changed, Update}
            ),
            gleam@otp@actor:continue(
                begin
                    _record = State,
                    {state,
                        erlang:element(2, _record),
                        erlang:element(3, _record),
                        erlang:element(4, _record),
                        none,
                        erlang:element(6, _record),
                        erlang:element(7, _record),
                        erlang:element(8, _record),
                        erlang:element(9, _record),
                        erlang:element(10, _record)}
                end
            );

        none ->
            gleam@otp@actor:continue(State)
    end.

-file("src/spoke.gleam", 685).
-spec handle_puback(state(), integer()) -> gleam@otp@actor:next(message(), state()).
handle_puback(State, Id) ->
    case spoke@internal@session:handle_puback(erlang:element(7, State), Id) of
        invalid_pub_ack_id ->
            drop_connection_and_notify(
                State,
                {some, <<"Received invalid packet id in publish ack"/utf8>>}
            );

        {publish_finished, Session} ->
            _pipe = begin
                _record = State,
                {state,
                    erlang:element(2, _record),
                    erlang:element(3, _record),
                    erlang:element(4, _record),
                    erlang:element(5, _record),
                    erlang:element(6, _record),
                    Session,
                    erlang:element(8, _record),
                    erlang:element(9, _record),
                    erlang:element(10, _record)}
            end,
            _pipe@1 = notify_if_publishes_finished(_pipe),
            gleam@otp@actor:continue(_pipe@1)
    end.

-file("src/spoke.gleam", 872).
-spec ok_or_drop_connection(
    state(),
    {ok, QRX} | {error, QRY},
    fun((QRY) -> binary()),
    fun((QRX) -> gleam@otp@actor:next(message(), state()))
) -> gleam@otp@actor:next(message(), state()).
ok_or_drop_connection(State, Result, Make_reason, Continuation) ->
    case Result of
        {ok, A} ->
            Continuation(A);

        {error, E} ->
            drop_connection_and_notify(State, {some, Make_reason(E)})
    end.

-file("src/spoke.gleam", 676).
-spec handle_unsuback(state(), integer()) -> gleam@otp@actor:next(message(), state()).
handle_unsuback(State, Id) ->
    Unsubs = erlang:element(9, State),
    ok_or_drop_connection(
        State,
        gleam_stdlib:map_get(Unsubs, Id),
        fun(_) -> <<"Received invalid packet id in unsubscribe ack"/utf8>> end,
        fun(Pending_unsub) ->
            gleam@erlang@process:send(Pending_unsub, {ok, nil}),
            gleam@otp@actor:continue(
                begin
                    _record = State,
                    {state,
                        erlang:element(2, _record),
                        erlang:element(3, _record),
                        erlang:element(4, _record),
                        erlang:element(5, _record),
                        erlang:element(6, _record),
                        erlang:element(7, _record),
                        erlang:element(8, _record),
                        gleam@dict:delete(Unsubs, Id),
                        erlang:element(10, _record)}
                end
            )
        end
    ).

-file("src/spoke.gleam", 888).
-spec to_packet_qos(qo_s()) -> spoke@internal@packet:qo_s().
to_packet_qos(Qos) ->
    case Qos of
        at_most_once ->
            qo_s0;

        at_least_once ->
            qo_s1;

        exactly_once ->
            qo_s2
    end.

-file("src/spoke.gleam", 513).
-spec handle_connect(state(), boolean(), gleam@option:option(publish_data())) -> gleam@otp@actor:next(message(), state()).
handle_connect(State, Clean_session, Will) ->
    gleam@bool:guard(
        gleam@option:is_some(erlang:element(5, State)),
        gleam@otp@actor:continue(State),
        fun() ->
            Auth@1 = gleam@option:map(
                erlang:element(3, erlang:element(3, State)),
                fun(Auth) ->
                    {auth_options,
                        erlang:element(2, Auth),
                        erlang:element(3, Auth)}
                end
            ),
            Will@2 = gleam@option:map(
                Will,
                fun(Will@1) ->
                    Data = {message_data,
                        erlang:element(2, Will@1),
                        erlang:element(3, Will@1),
                        erlang:element(5, Will@1)},
                    {Data, to_packet_qos(erlang:element(4, Will@1))}
                end
            ),
            Config = erlang:element(3, State),
            Conn = spoke@internal@connection:connect(
                erlang:element(6, Config),
                erlang:element(2, Config),
                Clean_session,
                erlang:element(4, Config),
                erlang:element(5, Config),
                Auth@1,
                Will@2
            ),
            Session = case Clean_session orelse spoke@internal@session:is_ephemeral(
                erlang:element(7, State)
            ) of
                true ->
                    spoke@internal@session:new(Clean_session);

                false ->
                    erlang:element(7, State)
            end,
            case Conn of
                {ok, Conn@1} ->
                    Selector = begin
                        _pipe = spoke@internal@connection:updates(Conn@1),
                        _pipe@1 = gleam_erlang_ffi:map_selector(
                            _pipe,
                            fun(Update) -> case Update of
                                    {received_packet, Packet} ->
                                        {process_received, Packet};

                                    {disconnected, Reason} ->
                                        {connection_dropped, Reason}
                                end end
                        ),
                        gleam@erlang@process:selecting(
                            _pipe@1,
                            erlang:element(2, State),
                            fun gleam@function:identity/1
                        )
                    end,
                    State@1 = begin
                        _record = State,
                        {state,
                            erlang:element(2, _record),
                            erlang:element(3, _record),
                            erlang:element(4, _record),
                            {some, Conn@1},
                            erlang:element(6, _record),
                            Session,
                            erlang:element(8, _record),
                            erlang:element(9, _record),
                            erlang:element(10, _record)}
                    end,
                    _pipe@2 = spoke@internal@session:packets_to_send_after_connect(
                        erlang:element(7, State@1)
                    ),
                    gleam@list:each(
                        _pipe@2,
                        fun(_capture) ->
                            spoke@internal@connection:send(Conn@1, _capture)
                        end
                    ),
                    gleam@otp@actor:with_selector(
                        gleam@otp@actor:continue(State@1),
                        Selector
                    );

                {error, _} ->
                    gleam@otp@actor:continue(
                        begin
                            _record@1 = State,
                            {state,
                                erlang:element(2, _record@1),
                                erlang:element(3, _record@1),
                                erlang:element(4, _record@1),
                                none,
                                erlang:element(6, _record@1),
                                erlang:element(7, _record@1),
                                erlang:element(8, _record@1),
                                erlang:element(9, _record@1),
                                erlang:element(10, _record@1)}
                        end
                    )
            end
        end
    ).

-file("src/spoke.gleam", 728).
-spec handle_subscribe(
    state(),
    list(subscribe_request()),
    gleam@erlang@process:subject({ok, list(subscription())} |
        {error, operation_error()})
) -> gleam@otp@actor:next(message(), state()).
handle_subscribe(State, Topics, Reply_to) ->
    guard_connected(
        State,
        fun() ->
            gleam@erlang@process:send(Reply_to, {error, not_connected}),
            gleam@otp@actor:continue(State)
        end,
        fun(Connection) ->
            {Session, Id} = spoke@internal@session:reserve_packet_id(
                erlang:element(7, State)
            ),
            Pending_sub = {pending_subscription, Topics, Reply_to},
            Pending_subs = begin
                _pipe = erlang:element(8, State),
                gleam@dict:insert(_pipe, Id, Pending_sub)
            end,
            Topics@1 = begin
                gleam@list:map(
                    Topics,
                    fun(Topic) ->
                        {subscribe_request,
                            erlang:element(2, Topic),
                            to_packet_qos(erlang:element(3, Topic))}
                    end
                )
            end,
            spoke@internal@connection:send(
                Connection,
                {subscribe, Id, Topics@1}
            ),
            gleam@otp@actor:continue(
                begin
                    _record = State,
                    {state,
                        erlang:element(2, _record),
                        erlang:element(3, _record),
                        erlang:element(4, _record),
                        erlang:element(5, _record),
                        erlang:element(6, _record),
                        Session,
                        Pending_subs,
                        erlang:element(9, _record),
                        erlang:element(10, _record)}
                end
            )
        end
    ).

-file("src/spoke.gleam", 896).
-spec from_packet_qos(spoke@internal@packet:qo_s()) -> qo_s().
from_packet_qos(Qos) ->
    case Qos of
        qo_s0 ->
            at_most_once;

        qo_s1 ->
            at_least_once;

        qo_s2 ->
            exactly_once
    end.

-file("src/spoke.gleam", 642).
-spec handle_suback(
    state(),
    integer(),
    list({ok, spoke@internal@packet:qo_s()} | {error, nil})
) -> gleam@otp@actor:next(message(), state()).
handle_suback(State, Id, Results) ->
    Subs = erlang:element(8, State),
    ok_or_drop_connection(
        State,
        gleam_stdlib:map_get(Subs, Id),
        fun(_) -> <<"Received invalid packet id in subscribe ack"/utf8>> end,
        fun(Pending_sub) ->
            ok_or_drop_connection(
                State,
                gleam@list:strict_zip(erlang:element(2, Pending_sub), Results),
                fun(_) ->
                    gleam@erlang@process:send(
                        erlang:element(3, Pending_sub),
                        {error, protocol_violation}
                    ),
                    <<"Received invalid number of results in subscribe ack"/utf8>>
                end,
                fun(Pairs) ->
                    Results@1 = begin
                        gleam@list:map(
                            Pairs,
                            fun(_use0) ->
                                {Topic, Result} = _use0,
                                _pipe = Result,
                                _pipe@1 = gleam@result:map(
                                    _pipe,
                                    fun(Qos) ->
                                        {successful_subscription,
                                            erlang:element(2, Topic),
                                            from_packet_qos(Qos)}
                                    end
                                ),
                                gleam@result:unwrap(
                                    _pipe@1,
                                    {failed_subscription,
                                        erlang:element(2, Topic)}
                                )
                            end
                        )
                    end,
                    gleam@erlang@process:send(
                        erlang:element(3, Pending_sub),
                        {ok, Results@1}
                    ),
                    gleam@otp@actor:continue(
                        begin
                            _record = State,
                            {state,
                                erlang:element(2, _record),
                                erlang:element(3, _record),
                                erlang:element(4, _record),
                                erlang:element(5, _record),
                                erlang:element(6, _record),
                                erlang:element(7, _record),
                                gleam@dict:delete(Subs, Id),
                                erlang:element(9, _record),
                                erlang:element(10, _record)}
                        end
                    )
                end
            )
        end
    ).

-file("src/spoke.gleam", 904).
-spec from_packet_connect_error(spoke@internal@packet:connect_error()) -> connect_error().
from_packet_connect_error(Error) ->
    case Error of
        bad_username_or_password ->
            bad_username_or_password;

        identifier_refused ->
            identifier_refused;

        not_authorized ->
            not_authorized;

        server_unavailable ->
            server_unavailable;

        unacceptable_protocol_version ->
            unacceptable_protocol_version
    end.

-file("src/spoke.gleam", 630).
-spec handle_connack(
    state(),
    {ok, boolean()} | {error, spoke@internal@packet:connect_error()}
) -> gleam@otp@actor:next(message(), state()).
handle_connack(State, Result) ->
    Update = case Result of
        {ok, Session_present} ->
            {connect_accepted, Session_present};

        {error, E} ->
            {connect_rejected, from_packet_connect_error(E)}
    end,
    gleam@erlang@process:send(
        erlang:element(4, State),
        {connection_state_changed, Update}
    ),
    gleam@otp@actor:continue(
        begin
            _record = State,
            {state,
                erlang:element(2, _record),
                erlang:element(3, _record),
                erlang:element(4, _record),
                erlang:element(5, _record),
                gleam@result:is_ok(Result),
                erlang:element(7, _record),
                erlang:element(8, _record),
                erlang:element(9, _record),
                erlang:element(10, _record)}
        end
    ).

-file("src/spoke.gleam", 611).
-spec process_packet(state(), spoke@internal@packet@client@incoming:packet()) -> gleam@otp@actor:next(message(), state()).
process_packet(State, Packet) ->
    case Packet of
        {conn_ack, Result} ->
            handle_connack(State, Result);

        {publish, Data} ->
            handle_incoming_publish(State, Data);

        {sub_ack, Id, Results} ->
            handle_suback(State, Id, Results);

        {unsub_ack, Id@1} ->
            handle_unsuback(State, Id@1);

        {pub_ack, Id@2} ->
            handle_puback(State, Id@2);

        {pub_rel, Id@3} ->
            handle_pubrel(State, Id@3);

        {pub_rec, Id@4} ->
            handle_pubrec(State, Id@4);

        {pub_comp, Id@5} ->
            handle_pubcomp(State, Id@5);

        ping_resp ->
            gleam@otp@actor:continue(State)
    end.

-file("src/spoke.gleam", 452).
-spec run_client(message(), state()) -> gleam@otp@actor:next(message(), state()).
run_client(Message, State) ->
    case Message of
        crash ->
            erlang:error(#{gleam_error => panic,
                    message => <<"Requested actor to crash"/utf8>>,
                    module => <<"spoke"/utf8>>,
                    function => <<"run_client"/utf8>>,
                    line => 454});

        {connect, Clean_session, Will} ->
            handle_connect(State, Clean_session, Will);

        {process_received, Packet} ->
            process_packet(State, Packet);

        {publish, Data} ->
            handle_outgoing_publish(State, Data);

        {subscribe, Topics, Reply_to} ->
            handle_subscribe(State, Topics, Reply_to);

        {unsubscribe, Topics@1, Reply_to@1} ->
            handle_unsubscribe(State, Topics@1, Reply_to@1);

        {connection_dropped, Reason} ->
            handle_connection_drop(State, Reason);

        {disconnect, Reply_to@2} ->
            handle_disconnect(State, Reply_to@2);

        {drop_connection_and_notify_client, Reason@1} ->
            drop_connection_and_notify(State, {some, Reason@1});

        {get_pending_publishes, Reply_to@3} ->
            get_pending_publishes(State, Reply_to@3);

        {wait_for_publishes_to_finish, Reply_to@4} ->
            handle_wait_for_publishes_to_finish(State, Reply_to@4)
    end.

-file("src/spoke.gleam", 342).
?DOC(false).
-spec start_session_with_ms_keep_alive(
    spoke@internal@session:session(),
    binary(),
    gleam@option:option(auth_details()),
    integer(),
    integer(),
    fun(() -> {ok,
            {fun((gleam@bytes_tree:bytes_tree()) -> {ok, nil} |
                    {error, binary()}),
                fun(() -> gleam@erlang@process:selector({ok, bitstring()} |
                    {error, binary()})),
                fun(() -> nil)}} |
        {error, binary()})
) -> client().
start_session_with_ms_keep_alive(
    Session,
    Client_id,
    Auth,
    Keep_alive_ms,
    Server_timeout_ms,
    Connector
) ->
    Connect = fun() ->
        gleam@result:'try'(
            Connector(),
            fun(_use0) ->
                {Send, Receive_next, Shutdown} = _use0,
                {ok, {byte_channel, Send, Receive_next, Shutdown}}
            end
        )
    end,
    Updates = gleam@erlang@process:new_subject(),
    Config = {config,
        Client_id,
        Auth,
        Keep_alive_ms,
        Server_timeout_ms,
        Connect},
    _assert_subject = gleam@otp@actor:start_spec(
        {spec,
            fun() -> init(Config, Session, Updates) end,
            100,
            fun run_client/2}
    ),
    {ok, Client} = case _assert_subject of
        {ok, _} -> _assert_subject;
        _assert_fail ->
            erlang:error(#{gleam_error => let_assert,
                        message => <<"Pattern match failed, no pattern matched the value."/utf8>>,
                        value => _assert_fail,
                        module => <<"spoke"/utf8>>,
                        function => <<"start_session_with_ms_keep_alive"/utf8>>,
                        line => 358})
    end,
    {client, Client, Updates, Config}.

-file("src/spoke.gleam", 201).
?DOC(
    " Starts a new MQTT session with the given options.\n"
    " Does not connect to the server, until `connect` is called.\n"
).
-spec start_session(connect_options()) -> client().
start_session(Connect_options) ->
    start_session_with_ms_keep_alive(
        spoke@internal@session:new(false),
        erlang:element(3, Connect_options),
        erlang:element(4, Connect_options),
        erlang:element(5, Connect_options) * 1000,
        erlang:element(6, Connect_options),
        erlang:element(2, Connect_options)
    ).

-file("src/spoke.gleam", 217).
?DOC(
    " Restores an MQTT session from a previously stored state,\n"
    " as returned by `disconnect`.\n"
    " Will return an error if the session state can not be restored\n"
    " from the provided string.\n"
    " Does not connect to the server, until `connect` is called.\n"
).
-spec restore_session(connect_options(), binary()) -> {ok, client()} |
    {error, binary()}.
restore_session(Connect_options, State) ->
    _pipe = spoke@internal@session:from_json(State),
    _pipe@1 = gleam@result:map_error(_pipe, fun gleam@string:inspect/1),
    gleam@result:map(
        _pipe@1,
        fun(_capture) ->
            start_session_with_ms_keep_alive(
                _capture,
                erlang:element(3, Connect_options),
                erlang:element(4, Connect_options),
                erlang:element(5, Connect_options) * 1000,
                erlang:element(6, Connect_options),
                erlang:element(2, Connect_options)
            )
        end
    ).
