-module(spoke@internal@connection).
-compile([no_auto_import, nowarn_unused_vars, nowarn_unused_function, nowarn_nomatch]).

-export([updates/1, send/2, shutdown/1, connect/7]).
-export_type([connection/0, disconnect/0, update/0, config/0, message/0, state/0, connection_state/0]).

-if(?OTP_RELEASE >= 27).
-define(MODULEDOC(Str), -moduledoc(Str)).
-define(DOC(Str), -doc(Str)).
-else.
-define(MODULEDOC(Str), -compile([])).
-define(DOC(Str), -compile([])).
-endif.

?MODULEDOC(false).

-opaque connection() :: {connection,
        gleam@erlang@process:subject(message()),
        gleam@erlang@process:selector(update())}.

-type disconnect() :: graceful_disconnect | {abnormal_disconnect, binary()}.

-type update() :: {disconnected, disconnect()} |
    {received_packet, spoke@internal@packet@client@incoming:packet()}.

-type config() :: {config,
        fun(() -> {ok, spoke@internal@transport:byte_channel()} |
            {error, binary()}),
        gleam@erlang@process:subject(update()),
        binary(),
        integer(),
        integer()}.

-type message() :: {send, spoke@internal@packet@client@outgoing:packet()} |
    {received,
        {bitstring(),
            {ok, list(spoke@internal@packet@client@incoming:packet())} |
                {error, binary()}}} |
    {process_received, spoke@internal@packet@client@incoming:packet()} |
    send_ping |
    {shut_down, gleam@option:option(binary())}.

-type state() :: {state,
        gleam@erlang@process:subject(message()),
        gleam@erlang@process:selector(message()),
        gleam@erlang@process:selector(message()),
        gleam@erlang@process:subject(update()),
        integer(),
        integer(),
        connection_state()}.

-type connection_state() :: {connecting_to_server,
        spoke@internal@transport:encoded_channel(),
        gleam@erlang@process:timer()} |
    {connected,
        spoke@internal@transport:encoded_channel(),
        gleam@erlang@process:timer(),
        gleam@option:option(gleam@erlang@process:timer())}.

-file("src/spoke/internal/connection.gleam", 104).
?DOC(false).
-spec updates(connection()) -> gleam@erlang@process:selector(update()).
updates(Connection) ->
    erlang:element(3, Connection).

-file("src/spoke/internal/connection.gleam", 109).
?DOC(false).
-spec send(connection(), spoke@internal@packet@client@outgoing:packet()) -> nil.
send(Connection, Packet) ->
    gleam@erlang@process:send(erlang:element(2, Connection), {send, Packet}).

-file("src/spoke/internal/connection.gleam", 114).
?DOC(false).
-spec shutdown(connection()) -> nil.
shutdown(Connection) ->
    gleam@erlang@process:send(erlang:element(2, Connection), {shut_down, none}).

-file("src/spoke/internal/connection.gleam", 284).
?DOC(false).
-spec send_to_session(state(), spoke@internal@packet@client@incoming:packet()) -> state().
send_to_session(State, Packet) ->
    gleam@erlang@process:send(
        erlang:element(5, State),
        {received_packet, Packet}
    ),
    State.

-file("src/spoke/internal/connection.gleam", 367).
?DOC(false).
-spec next_selector(
    spoke@internal@transport:encoded_channel(),
    bitstring(),
    gleam@erlang@process:selector(message())
) -> gleam@erlang@process:selector(message()).
next_selector(Channel, Channel_state, Base) ->
    _pipe = gleam_erlang_ffi:map_selector(
        (erlang:element(3, Channel))(Channel_state),
        fun(_capture) -> {received, _capture} end
    ),
    gleam_erlang_ffi:merge_selector(_pipe, Base).

-file("src/spoke/internal/connection.gleam", 417).
?DOC(false).
-spec start_ping_timeout(state()) -> gleam@erlang@process:timer().
start_ping_timeout(State) ->
    gleam@erlang@process:send_after(
        erlang:element(2, State),
        erlang:element(6, State),
        send_ping
    ).

-file("src/spoke/internal/connection.gleam", 401).
?DOC(false).
-spec send_packet(state(), spoke@internal@packet@client@outgoing:packet()) -> {ok,
        state()} |
    {error, binary()}.
send_packet(State, Packet) ->
    case erlang:element(8, State) of
        {connecting_to_server, Channel, _} ->
            gleam@result:'try'(
                (erlang:element(2, Channel))(Packet),
                fun(_) -> {ok, State} end
            );

        {connected, Channel@1, Ping, Disconnect} ->
            gleam@result:'try'(
                (erlang:element(2, Channel@1))(Packet),
                fun(_) ->
                    gleam@erlang@process:cancel_timer(Ping),
                    Ping@1 = start_ping_timeout(State),
                    {ok,
                        begin
                            _record = State,
                            {state,
                                erlang:element(2, _record),
                                erlang:element(3, _record),
                                erlang:element(4, _record),
                                erlang:element(5, _record),
                                erlang:element(6, _record),
                                erlang:element(7, _record),
                                {connected, Channel@1, Ping@1, Disconnect}}
                        end}
                end
            )
    end.

-file("src/spoke/internal/connection.gleam", 421).
?DOC(false).
-spec get_channel(state()) -> spoke@internal@transport:encoded_channel().
get_channel(State) ->
    case erlang:element(8, State) of
        {connected, Channel, _, _} ->
            Channel;

        {connecting_to_server, Channel@1, _} ->
            Channel@1
    end.

-file("src/spoke/internal/connection.gleam", 432).
?DOC(false).
-spec stop(gleam@option:option(binary())) -> any().
stop(Reason) ->
    case Reason of
        none ->
            gleam@erlang@process:send_exit(erlang:self());

        {some, Reason@1} ->
            gleam@erlang@process:send_abnormal_exit(erlang:self(), Reason@1)
    end,
    erlang:error(#{gleam_error => panic,
            message => <<"We should be dead by now"/utf8>>,
            module => <<"spoke/internal/connection"/utf8>>,
            function => <<"stop"/utf8>>,
            line => 438}).

-file("src/spoke/internal/connection.gleam", 245).
?DOC(false).
-spec shut_down(state(), gleam@option:option(binary())) -> state().
shut_down(State, Reason) ->
    (erlang:element(4, get_channel(State)))(),
    stop(Reason).

-file("src/spoke/internal/connection.gleam", 289).
?DOC(false).
-spec handle_connack(
    state(),
    {ok, boolean()} | {error, spoke@internal@packet:connect_error()}
) -> state().
handle_connack(State, Status) ->
    case erlang:element(8, State) of
        {connecting_to_server, Channel, Disconnect_timer} ->
            gleam@erlang@process:cancel_timer(Disconnect_timer),
            gleam@erlang@process:send(
                erlang:element(5, State),
                {received_packet, {conn_ack, Status}}
            ),
            case Status of
                {ok, _} ->
                    Connection_state = {connected,
                        Channel,
                        start_ping_timeout(State),
                        none},
                    _record = State,
                    {state,
                        erlang:element(2, _record),
                        erlang:element(3, _record),
                        erlang:element(4, _record),
                        erlang:element(5, _record),
                        erlang:element(6, _record),
                        erlang:element(7, _record),
                        Connection_state};

                {error, _} ->
                    shut_down(State, none)
            end;

        {connected, _, _, _} ->
            shut_down(
                State,
                {some, <<"Got ConnAck while already connected"/utf8>>}
            )
    end.

-file("src/spoke/internal/connection.gleam", 428).
?DOC(false).
-spec stop_abnormal(binary()) -> any().
stop_abnormal(Reason) ->
    stop({some, Reason}).

-file("src/spoke/internal/connection.gleam", 220).
?DOC(false).
-spec ok_or_exit(
    {ok, QAU} | {error, QAV},
    fun((QAV) -> binary()),
    fun((QAU) -> QAY)
) -> QAY.
ok_or_exit(Result, Make_reason, Continuation) ->
    case Result of
        {ok, A} ->
            Continuation(A);

        {error, E} ->
            stop_abnormal(Make_reason(E))
    end.

-file("src/spoke/internal/connection.gleam", 256).
?DOC(false).
-spec handle_receive(
    state(),
    bitstring(),
    {ok, list(spoke@internal@packet@client@incoming:packet())} |
        {error, binary()}
) -> state().
handle_receive(State, New_conn_state, Packets) ->
    ok_or_exit(
        Packets,
        fun(E) ->
            <<"Transport channel error while receiving: "/utf8, E/binary>>
        end,
        fun(Packets@1) ->
            gleam@list:each(
                Packets@1,
                fun(Packet) ->
                    gleam@erlang@process:send(
                        erlang:element(2, State),
                        {process_received, Packet}
                    )
                end
            ),
            Selector = next_selector(
                get_channel(State),
                New_conn_state,
                erlang:element(3, State)
            ),
            _record = State,
            {state,
                erlang:element(2, _record),
                erlang:element(3, _record),
                Selector,
                erlang:element(5, _record),
                erlang:element(6, _record),
                erlang:element(7, _record),
                erlang:element(8, _record)}
        end
    ).

-file("src/spoke/internal/connection.gleam", 354).
?DOC(false).
-spec handle_pingresp(state()) -> state().
handle_pingresp(State) ->
    case erlang:element(8, State) of
        {connected, Channel, Ping, {some, Shutdown_timer}} ->
            gleam@erlang@process:cancel_timer(Shutdown_timer),
            _record = State,
            {state,
                erlang:element(2, _record),
                erlang:element(3, _record),
                erlang:element(4, _record),
                erlang:element(5, _record),
                erlang:element(6, _record),
                erlang:element(7, _record),
                {connected, Channel, Ping, none}};

        {connected, _, _, none} ->
            stop_abnormal(
                <<"No shutdown timer active when receiving PingResp"/utf8>>
            );

        {connecting_to_server, _, _} ->
            stop_abnormal(<<"Got unexpected PingResp"/utf8>>)
    end.

-file("src/spoke/internal/connection.gleam", 276).
?DOC(false).
-spec process_packet(state(), spoke@internal@packet@client@incoming:packet()) -> state().
process_packet(State, Packet) ->
    case Packet of
        {conn_ack, Data} ->
            handle_connack(State, Data);

        ping_resp ->
            handle_pingresp(State);

        Packet@1 ->
            send_to_session(State, Packet@1)
    end.

-file("src/spoke/internal/connection.gleam", 376).
?DOC(false).
-spec send_packet_or_stop(
    state(),
    spoke@internal@packet@client@outgoing:packet(),
    fun((state()) -> state())
) -> state().
send_packet_or_stop(State, Packet, If_success) ->
    ok_or_exit(
        send_packet(State, Packet),
        fun(E) ->
            <<<<<<"Transport channel error '"/utf8, E/binary>>/binary,
                    "' while sending packet "/utf8>>/binary,
                (gleam@string:inspect(Packet))/binary>>
        end,
        fun(State@1) -> If_success(State@1) end
    ).

-file("src/spoke/internal/connection.gleam", 251).
?DOC(false).
-spec send_from_session(state(), spoke@internal@packet@client@outgoing:packet()) -> state().
send_from_session(State, Packet) ->
    send_packet_or_stop(State, Packet, fun(State@1) -> State@1 end).

-file("src/spoke/internal/connection.gleam", 316).
?DOC(false).
-spec send_ping(state()) -> state().
send_ping(State) ->
    case erlang:element(8, State) of
        {connected, Channel, Ping, Disconnect} ->
            case {Disconnect, gleam@erlang@process:cancel_timer(Ping)} of
                {_, {cancelled, Time_remaining}} ->
                    stop_abnormal(
                        <<"A ping scheduling overlap happened, previous time remaining: "/utf8,
                            (gleam@string:inspect(Time_remaining))/binary>>
                    );

                {{some, _}, _} ->
                    stop_abnormal(
                        <<"A ping was scheduled while waiting for a ping response"/utf8>>
                    );

                {none, _} ->
                    send_packet_or_stop(
                        State,
                        ping_req,
                        fun(State@1) ->
                            Disconnect@1 = gleam@erlang@process:send_after(
                                erlang:element(2, State@1),
                                erlang:element(7, State@1),
                                {shut_down, {some, <<"Ping timeout"/utf8>>}}
                            ),
                            _record = State@1,
                            {state,
                                erlang:element(2, _record),
                                erlang:element(3, _record),
                                erlang:element(4, _record),
                                erlang:element(5, _record),
                                erlang:element(6, _record),
                                erlang:element(7, _record),
                                {connected, Channel, Ping, {some, Disconnect@1}}}
                        end
                    )
            end;

        {connecting_to_server, _, _} ->
            stop_abnormal(
                <<"Trying to send ping while still connecting to server"/utf8>>
            )
    end.

-file("src/spoke/internal/connection.gleam", 231).
?DOC(false).
-spec run(state()) -> nil.
run(State) ->
    Message = gleam_erlang_ffi:select(erlang:element(4, State)),
    State@1 = case Message of
        {received, {New_channel_state, Packets}} ->
            handle_receive(State, New_channel_state, Packets);

        {process_received, Packet} ->
            process_packet(State, Packet);

        {send, Packet@1} ->
            send_from_session(State, Packet@1);

        send_ping ->
            send_ping(State);

        {shut_down, Reason} ->
            shut_down(State, Reason)
    end,
    run(State@1).

-file("src/spoke/internal/connection.gleam", 162).
?DOC(false).
-spec establish_channel(
    gleam@erlang@process:pid_(),
    gleam@erlang@process:subject(message()),
    gleam@erlang@process:subject(config()),
    boolean(),
    gleam@option:option(spoke@internal@packet:auth_options()),
    gleam@option:option({spoke@internal@packet:message_data(),
        spoke@internal@packet:qo_s()})
) -> nil.
establish_channel(Parent, Mailbox, Connect, Clean_session, Auth, Will) ->
    _assert_subject = gleam_erlang_ffi:'receive'(Connect, 1000),
    {ok, Config} = case _assert_subject of
        {ok, _} -> _assert_subject;
        _assert_fail ->
            erlang:error(#{gleam_error => let_assert,
                        message => <<"Pattern match failed, no pattern matched the value."/utf8>>,
                        value => _assert_fail,
                        module => <<"spoke/internal/connection"/utf8>>,
                        function => <<"establish_channel"/utf8>>,
                        line => 171})
    end,
    ok_or_exit(
        (erlang:element(2, Config))(),
        fun(E) ->
            <<"Failed to establish connection to server: "/utf8, E/binary>>
        end,
        fun(Channel) ->
            Channel@1 = spoke@internal@transport:encode_channel(Channel),
            Disconnect_timer = gleam@erlang@process:send_after(
                Mailbox,
                erlang:element(6, Config),
                {shut_down, {some, <<"Connect timed out"/utf8>>}}
            ),
            Connection_state = {connecting_to_server,
                Channel@1,
                Disconnect_timer},
            Parent_monitor = gleam@erlang@process:monitor_process(Parent),
            Base_selector = begin
                _pipe = gleam_erlang_ffi:new_selector(),
                _pipe@1 = gleam@erlang@process:selecting_process_down(
                    _pipe,
                    Parent_monitor,
                    fun(_) -> {shut_down, none} end
                ),
                gleam@erlang@process:selecting(
                    _pipe@1,
                    Mailbox,
                    fun gleam@function:identity/1
                )
            end,
            State = {state,
                Mailbox,
                Base_selector,
                next_selector(Channel@1, <<>>, Base_selector),
                erlang:element(3, Config),
                erlang:element(5, Config),
                erlang:element(6, Config),
                Connection_state},
            Connect_options = {connect_options,
                Clean_session,
                erlang:element(4, Config),
                erlang:element(5, Config) div 1000,
                Auth,
                Will},
            ok_or_exit(
                send_packet(State, {connect, Connect_options}),
                fun(E@1) ->
                    <<"Error sending connect packet to server: "/utf8,
                        E@1/binary>>
                end,
                fun(State@1) -> run(State@1) end
            )
        end
    ).

-file("src/spoke/internal/connection.gleam", 46).
?DOC(false).
-spec connect(
    fun(() -> {ok, spoke@internal@transport:byte_channel()} | {error, binary()}),
    binary(),
    boolean(),
    integer(),
    integer(),
    gleam@option:option(spoke@internal@packet:auth_options()),
    gleam@option:option({spoke@internal@packet:message_data(),
        spoke@internal@packet:qo_s()})
) -> {ok, connection()} | {error, binary()}.
connect(
    Create_channel,
    Client_id,
    Clean_session,
    Keep_alive_ms,
    Server_timeout_ms,
    Auth,
    Will
) ->
    Ack_subject = gleam@erlang@process:new_subject(),
    Parent = erlang:self(),
    Child = gleam@erlang@process:start(
        fun() ->
            Messages = gleam@erlang@process:new_subject(),
            Connect = gleam@erlang@process:new_subject(),
            gleam@erlang@process:send(Ack_subject, {Messages, Connect}),
            establish_channel(
                Parent,
                Messages,
                Connect,
                Clean_session,
                Auth,
                Will
            )
        end,
        true
    ),
    Monitor = gleam@erlang@process:monitor_process(Child),
    gleam@erlang@process:unlink(Child),
    Updates = gleam@erlang@process:new_subject(),
    Config = {config,
        Create_channel,
        Updates,
        Client_id,
        Keep_alive_ms,
        Server_timeout_ms},
    gleam@result:'try'(
        begin
            _pipe = gleam_erlang_ffi:'receive'(Ack_subject, Server_timeout_ms),
            gleam@result:map_error(
                _pipe,
                fun(_) ->
                    gleam@erlang@process:kill(Child),
                    <<"Starting connection process timed out"/utf8>>
                end
            )
        end,
        fun(_use0) ->
            {Messages@1, Connect@1} = _use0,
            gleam@erlang@process:send(Connect@1, Config),
            Updates@1 = begin
                _pipe@1 = gleam_erlang_ffi:new_selector(),
                _pipe@2 = gleam@erlang@process:selecting(
                    _pipe@1,
                    Updates,
                    fun gleam@function:identity/1
                ),
                gleam@erlang@process:selecting_process_down(
                    _pipe@2,
                    Monitor,
                    fun(Down) ->
                        Reason = erlang:element(3, Down),
                        Normal = gleam_stdlib:identity(normal),
                        Killed = gleam_stdlib:identity(killed),
                        Kind = begin
                            gleam@bool:guard(
                                Reason =:= Normal,
                                graceful_disconnect,
                                fun() ->
                                    gleam@bool:guard(
                                        Reason =:= Killed,
                                        {abnormal_disconnect, <<"Killed"/utf8>>},
                                        fun() ->
                                            {abnormal_disconnect,
                                                gleam@string:inspect(Reason)}
                                        end
                                    )
                                end
                            )
                        end,
                        {disconnected, Kind}
                    end
                )
            end,
            {ok, {connection, Messages@1, Updates@1}}
        end
    ).
