-module(spoke@internal@packet@decode).
-compile([no_auto_import, nowarn_unused_vars, nowarn_unused_function, nowarn_nomatch]).

-export([string/1, integer/1, varint/1, zero_length/3, only_packet_id/4, strings/2, unsubscribe/3, connack/3, suback/3, connect/3, publish/3, subscribe/3]).
-export_type([decode_error/0]).

-if(?OTP_RELEASE >= 27).
-define(MODULEDOC(Str), -moduledoc(Str)).
-define(DOC(Str), -doc(Str)).
-else.
-define(MODULEDOC(Str), -compile([])).
-define(DOC(Str), -compile([])).
-endif.

?MODULEDOC(false).

-type decode_error() :: {invalid_packet_identifier, integer()} |
    data_too_short |
    invalid_data |
    invalid_utf8 |
    invalid_qo_s |
    var_int_too_large.

-file("src/spoke/internal/packet/decode.gleam", 253).
?DOC(false).
-spec string(bitstring()) -> {ok, {binary(), bitstring()}} |
    {error, decode_error()}.
string(Bits) ->
    case Bits of
        <<Len:16/big-unsigned, Bytes:Len/binary, Rest/bitstring>> ->
            gleam@result:'try'(
                begin
                    _pipe = gleam@bit_array:to_string(Bytes),
                    gleam@result:map_error(_pipe, fun(_) -> invalid_utf8 end)
                end,
                fun(Str) -> {ok, {Str, Rest}} end
            );

        _ ->
            {error, data_too_short}
    end.

-file("src/spoke/internal/packet/decode.gleam", 267).
?DOC(false).
-spec integer(bitstring()) -> {ok, {integer(), bitstring()}} |
    {error, decode_error()}.
integer(Bits) ->
    case Bits of
        <<Val:16/big-unsigned, Rest/binary>> ->
            {ok, {Val, Rest}};

        _ ->
            {error, data_too_short}
    end.

-file("src/spoke/internal/packet/decode.gleam", 278).
?DOC(false).
-spec accumulate_varint(integer(), integer(), bitstring()) -> {ok,
        {integer(), bitstring()}} |
    {error, decode_error()}.
accumulate_varint(Value, Multiplier, Bytes) ->
    case Bytes of
        _ when Multiplier >= 268435456 ->
            {error, var_int_too_large};

        <<Continue:1, Next:7, Rest/bitstring>> ->
            Value@1 = Value + (Multiplier * Next),
            case Continue of
                0 ->
                    {ok, {Value@1, Rest}};

                _ ->
                    accumulate_varint(Value@1, 128 * Multiplier, Rest)
            end;

        _ ->
            {error, data_too_short}
    end.

-file("src/spoke/internal/packet/decode.gleam", 274).
?DOC(false).
-spec varint(bitstring()) -> {ok, {integer(), bitstring()}} |
    {error, decode_error()}.
varint(Bytes) ->
    accumulate_varint(0, 1, Bytes).

-file("src/spoke/internal/packet/decode.gleam", 297).
?DOC(false).
-spec split_fixed_data(bitstring(), integer()) -> {ok,
        {bitstring(), bitstring()}} |
    {error, decode_error()}.
split_fixed_data(Bytes, Len) ->
    case Bytes of
        <<Data:Len/binary, Rest/binary>> ->
            {ok, {Data, Rest}};

        _ ->
            {error, data_too_short}
    end.

-file("src/spoke/internal/packet/decode.gleam", 190).
?DOC(false).
-spec zero_length(bitstring(), bitstring(), OYY) -> {ok, {OYY, bitstring()}} |
    {error, decode_error()}.
zero_length(Flags, Data, Value) ->
    gleam@result:'try'(
        split_fixed_data(Data, 1),
        fun(_use0) ->
            {Data@1, Rest} = _use0,
            case {Flags, Data@1} of
                {<<0:4>>, <<0:8>>} ->
                    {ok, {Value, Rest}};

                {_, _} ->
                    {error, invalid_data}
            end
        end
    ).

-file("src/spoke/internal/packet/decode.gleam", 308).
?DOC(false).
-spec split_var_data(bitstring()) -> {ok, {bitstring(), bitstring()}} |
    {error, decode_error()}.
split_var_data(Bytes) ->
    gleam@result:'try'(
        varint(Bytes),
        fun(_use0) ->
            {Len, Rest} = _use0,
            split_fixed_data(Rest, Len)
        end
    ).

-file("src/spoke/internal/packet/decode.gleam", 313).
?DOC(false).
-spec must_equal(OZY, OZY) -> {ok, nil} | {error, decode_error()}.
must_equal(Input, Expected) ->
    case Input =:= Expected of
        true ->
            {ok, nil};

        false ->
            {error, invalid_data}
    end.

-file("src/spoke/internal/packet/decode.gleam", 219).
?DOC(false).
-spec only_packet_id(
    bitstring(),
    bitstring(),
    integer(),
    fun((integer()) -> OZF)
) -> {ok, {OZF, bitstring()}} | {error, decode_error()}.
only_packet_id(Flags, Data, Expected_flags, Construct) ->
    case Flags of
        <<Received_flags:4>> ->
            gleam@result:'try'(
                begin
                    _pipe = Received_flags,
                    must_equal(_pipe, Expected_flags)
                end,
                fun(_) ->
                    gleam@result:'try'(
                        split_var_data(Data),
                        fun(_use0) ->
                            {Data@1, Remainder} = _use0,
                            case Data@1 of
                                <<Packet_id:16/big-unsigned>> ->
                                    {ok, {Construct(Packet_id), Remainder}};

                                _ ->
                                    {error, invalid_data}
                            end
                        end
                    )
                end
            );

        _ ->
            {error, invalid_data}
    end.

-file("src/spoke/internal/packet/decode.gleam", 320).
?DOC(false).
-spec must_be_long_enough({ok, PAB} | {error, decode_error()}) -> {ok, PAB} |
    {error, decode_error()}.
must_be_long_enough(Result) ->
    case Result of
        {ok, Val} ->
            {ok, Val};

        {error, data_too_short} ->
            {error, invalid_data};

        {error, E} ->
            {error, E}
    end.

-file("src/spoke/internal/packet/decode.gleam", 240).
?DOC(false).
-spec strings(bitstring(), list(binary())) -> {ok, list(binary())} |
    {error, decode_error()}.
strings(Bits, Results) ->
    case Bits of
        <<>> ->
            {ok, lists:reverse(Results)};

        Bits@1 ->
            gleam@result:'try'(
                begin
                    _pipe = string(Bits@1),
                    must_be_long_enough(_pipe)
                end,
                fun(_use0) ->
                    {Str, Rest} = _use0,
                    _pipe@1 = Rest,
                    strings(_pipe@1, [Str | Results])
                end
            )
    end.

-file("src/spoke/internal/packet/decode.gleam", 134).
?DOC(false).
-spec unsubscribe(
    bitstring(),
    bitstring(),
    fun((integer(), list(binary())) -> OYS)
) -> {ok, {OYS, bitstring()}} | {error, decode_error()}.
unsubscribe(Flags, Data, Construct) ->
    gleam@result:'try'(
        begin
            _pipe = Flags,
            must_equal(_pipe, <<2:4>>)
        end,
        fun(_) ->
            gleam@result:'try'(
                split_var_data(Data),
                fun(_use0) ->
                    {Data@1, Remainder} = _use0,
                    gleam@result:'try'(
                        integer(Data@1),
                        fun(_use0@1) ->
                            {Packet_id, Rest} = _use0@1,
                            gleam@result:'try'(
                                strings(Rest, []),
                                fun(Topics) ->
                                    {ok,
                                        {Construct(Packet_id, Topics),
                                            Remainder}}
                                end
                            )
                        end
                    )
                end
            )
        end
    ).

-file("src/spoke/internal/packet/decode.gleam", 328).
?DOC(false).
-spec decode_connack_code(integer(), integer()) -> {ok,
        {ok, boolean()} | {error, spoke@internal@packet:connect_error()}} |
    {error, decode_error()}.
decode_connack_code(Session_present, Code) ->
    case Code of
        0 ->
            {ok, {ok, Session_present =:= 1}};

        1 ->
            {ok, {error, unacceptable_protocol_version}};

        2 ->
            {ok, {error, identifier_refused}};

        3 ->
            {ok, {error, server_unavailable}};

        4 ->
            {ok, {error, bad_username_or_password}};

        5 ->
            {ok, {error, not_authorized}};

        _ ->
            {error, invalid_data}
    end.

-file("src/spoke/internal/packet/decode.gleam", 100).
?DOC(false).
-spec connack(
    bitstring(),
    bitstring(),
    fun(({ok, boolean()} | {error, spoke@internal@packet:connect_error()}) -> OYK)
) -> {ok, {OYK, bitstring()}} | {error, decode_error()}.
connack(Flags, Data, Construct) ->
    gleam@result:'try'(
        split_fixed_data(Data, 3),
        fun(_use0) ->
            {Data@1, Rest} = _use0,
            case {Flags, Data@1} of
                {<<0:4>>, <<2:8, 0:7, Session_present:1, Return_code:8>>} ->
                    gleam@result:'try'(
                        decode_connack_code(Session_present, Return_code),
                        fun(Status) ->
                            Result = Construct(Status),
                            {ok, {Result, Rest}}
                        end
                    );

                {_, _} ->
                    {error, invalid_data}
            end
        end
    ).

-file("src/spoke/internal/packet/decode.gleam", 357).
?DOC(false).
-spec decode_suback_return(integer()) -> {ok,
        {ok, spoke@internal@packet:qo_s()} | {error, nil}} |
    {error, decode_error()}.
decode_suback_return(Val) ->
    case Val of
        0 ->
            {ok, {ok, qo_s0}};

        1 ->
            {ok, {ok, qo_s1}};

        2 ->
            {ok, {ok, qo_s2}};

        16#80 ->
            {ok, {error, nil}};

        _ ->
            {error, invalid_data}
    end.

-file("src/spoke/internal/packet/decode.gleam", 343).
?DOC(false).
-spec decode_suback_returns(
    bitstring(),
    list({ok, spoke@internal@packet:qo_s()} | {error, nil})
) -> {ok, list({ok, spoke@internal@packet:qo_s()} | {error, nil})} |
    {error, decode_error()}.
decode_suback_returns(Bytes, Codes) ->
    case Bytes of
        <<>> ->
            {ok, lists:reverse(Codes)};

        <<Val:8, Rest/binary>> ->
            gleam@result:'try'(
                decode_suback_return(Val),
                fun(Code) -> _pipe = Rest,
                    decode_suback_returns(_pipe, [Code | Codes]) end
            );

        _ ->
            {error, invalid_data}
    end.

-file("src/spoke/internal/packet/decode.gleam", 202).
?DOC(false).
-spec suback(
    bitstring(),
    bitstring(),
    fun((integer(), list({ok, spoke@internal@packet:qo_s()} | {error, nil})) -> OZC)
) -> {ok, {OZC, bitstring()}} | {error, decode_error()}.
suback(Flags, Data, Construct) ->
    case {Flags, Data} of
        {<<0:4>>, _} ->
            gleam@result:'try'(
                split_var_data(Data),
                fun(_use0) ->
                    {Data@1, Remainder} = _use0,
                    gleam@result:'try'(
                        integer(Data@1),
                        fun(_use0@1) ->
                            {Packet_id, Rest} = _use0@1,
                            gleam@result:'try'(
                                decode_suback_returns(Rest, []),
                                fun(Return_codes) ->
                                    {ok,
                                        {Construct(Packet_id, Return_codes),
                                            Remainder}}
                                end
                            )
                        end
                    )
                end
            );

        {_, _} ->
            {error, invalid_data}
    end.

-file("src/spoke/internal/packet/decode.gleam", 394).
?DOC(false).
-spec will(bitstring(), integer()) -> {ok,
        {spoke@internal@packet:message_data(), bitstring()}} |
    {error, decode_error()}.
will(Bytes, Retain) ->
    gleam@result:'try'(
        string(Bytes),
        fun(_use0) ->
            {Topic, Rest} = _use0,
            gleam@result:'try'(
                integer(Rest),
                fun(_use0@1) ->
                    {Payload_len, Rest@1} = _use0@1,
                    gleam@result:'try'(
                        split_fixed_data(Rest@1, Payload_len),
                        fun(_use0@2) ->
                            {Payload, Rest@2} = _use0@2,
                            Will = {message_data, Topic, Payload, Retain =:= 1},
                            {ok, {Will, Rest@2}}
                        end
                    )
                end
            )
        end
    ).

-file("src/spoke/internal/packet/decode.gleam", 406).
?DOC(false).
-spec auth_options(bitstring(), boolean()) -> {ok,
        {spoke@internal@packet:auth_options(), bitstring()}} |
    {error, decode_error()}.
auth_options(Bytes, Has_password) ->
    gleam@result:'try'(
        string(Bytes),
        fun(_use0) ->
            {Username, Rest} = _use0,
            gleam@result:'try'(case Has_password of
                    true ->
                        gleam@result:'try'(
                            integer(Rest),
                            fun(_use0@1) ->
                                {Len, Rest@1} = _use0@1,
                                case Rest@1 of
                                    <<Password:Len/binary, Rest@2/bitstring>> ->
                                        {ok, {{some, Password}, Rest@2}};

                                    _ ->
                                        {error, invalid_data}
                                end
                            end
                        );

                    false ->
                        {ok, {none, Rest}}
                end, fun(_use0@2) ->
                    {Password@1, Rest@3} = _use0@2,
                    Auth = {auth_options, Username, Password@1},
                    {ok, {Auth, Rest@3}}
                end)
        end
    ).

-file("src/spoke/internal/packet/decode.gleam", 426).
?DOC(false).
-spec decode_qos(integer()) -> {ok, spoke@internal@packet:qo_s()} |
    {error, decode_error()}.
decode_qos(Val) ->
    case Val of
        0 ->
            {ok, qo_s0};

        1 ->
            {ok, qo_s1};

        2 ->
            {ok, qo_s2};

        _ ->
            {error, invalid_qo_s}
    end.

-file("src/spoke/internal/packet/decode.gleam", 22).
?DOC(false).
-spec connect(
    bitstring(),
    bitstring(),
    fun((spoke@internal@packet:connect_options()) -> OYH)
) -> {ok, {OYH, bitstring()}} | {error, decode_error()}.
connect(Flags, Data, Construct) ->
    gleam@result:'try'(
        begin
            _pipe = Flags,
            must_equal(_pipe, <<0:4>>)
        end,
        fun(_) ->
            gleam@result:'try'(
                split_var_data(Data),
                fun(_use0) ->
                    {Data@1, Remainder} = _use0,
                    gleam@result:'try'(
                        string(Data@1),
                        fun(_use0@1) ->
                            {Name, Rest} = _use0@1,
                            gleam@result:'try'(
                                begin
                                    _pipe@1 = Name,
                                    must_equal(_pipe@1, <<"MQTT"/utf8>>)
                                end,
                                fun(_) -> gleam@result:'try'(case Rest of
                                            <<4:8, Rest@1/bitstring>> ->
                                                {ok, Rest@1};

                                            _ ->
                                                {error, invalid_data}
                                        end, fun(Rest@2) ->
                                            gleam@result:'try'(
                                                split_fixed_data(Rest@2, 1),
                                                fun(_use0@2) ->
                                                    {Connect_flags, Rest@3} = _use0@2,
                                                    gleam@result:'try'(
                                                        integer(Rest@3),
                                                        fun(_use0@3) ->
                                                            {Keep_alive, Rest@4} = _use0@3,
                                                            case Connect_flags of
                                                                <<Uname_flag:1,
                                                                    Pw_flag:1,
                                                                    Will_retain:1,
                                                                    Will_qos:2,
                                                                    Will_flag:1,
                                                                    Clean_session:1,
                                                                    0:1>> ->
                                                                    gleam@result:'try'(
                                                                        string(
                                                                            Rest@4
                                                                        ),
                                                                        fun(
                                                                            _use0@4
                                                                        ) ->
                                                                            {Client_id,
                                                                                Rest@5} = _use0@4,
                                                                            gleam@result:'try'(
                                                                                case Will_flag of
                                                                                    1 ->
                                                                                        gleam@result:'try'(
                                                                                            decode_qos(
                                                                                                Will_qos
                                                                                            ),
                                                                                            fun(
                                                                                                Qos
                                                                                            ) ->
                                                                                                gleam@result:'try'(
                                                                                                    will(
                                                                                                        Rest@5,
                                                                                                        Will_retain
                                                                                                    ),
                                                                                                    fun(
                                                                                                        _use0@5
                                                                                                    ) ->
                                                                                                        {Will,
                                                                                                            Rest@6} = _use0@5,
                                                                                                        {ok,
                                                                                                            {{some,
                                                                                                                    {Will,
                                                                                                                        Qos}},
                                                                                                                Rest@6}}
                                                                                                    end
                                                                                                )
                                                                                            end
                                                                                        );

                                                                                    _ ->
                                                                                        gleam@result:'try'(
                                                                                            begin
                                                                                                _pipe@2 = Will_retain,
                                                                                                must_equal(
                                                                                                    _pipe@2,
                                                                                                    0
                                                                                                )
                                                                                            end,
                                                                                            fun(
                                                                                                _
                                                                                            ) ->
                                                                                                gleam@result:'try'(
                                                                                                    begin
                                                                                                        _pipe@3 = Will_qos,
                                                                                                        must_equal(
                                                                                                            _pipe@3,
                                                                                                            0
                                                                                                        )
                                                                                                    end,
                                                                                                    fun(
                                                                                                        _
                                                                                                    ) ->
                                                                                                        {ok,
                                                                                                            {none,
                                                                                                                Rest@5}}
                                                                                                    end
                                                                                                )
                                                                                            end
                                                                                        )
                                                                                end,
                                                                                fun(
                                                                                    _use0@6
                                                                                ) ->
                                                                                    {Will@1,
                                                                                        Rest@7} = _use0@6,
                                                                                    gleam@result:'try'(
                                                                                        case Uname_flag of
                                                                                            1 ->
                                                                                                gleam@result:'try'(
                                                                                                    auth_options(
                                                                                                        Rest@7,
                                                                                                        Pw_flag
                                                                                                        =:= 1
                                                                                                    ),
                                                                                                    fun(
                                                                                                        _use0@7
                                                                                                    ) ->
                                                                                                        {Auth,
                                                                                                            Rest@8} = _use0@7,
                                                                                                        {ok,
                                                                                                            {{some,
                                                                                                                    Auth},
                                                                                                                Rest@8}}
                                                                                                    end
                                                                                                );

                                                                                            _ ->
                                                                                                gleam@result:'try'(
                                                                                                    begin
                                                                                                        _pipe@4 = Pw_flag,
                                                                                                        must_equal(
                                                                                                            _pipe@4,
                                                                                                            0
                                                                                                        )
                                                                                                    end,
                                                                                                    fun(
                                                                                                        _
                                                                                                    ) ->
                                                                                                        {ok,
                                                                                                            {none,
                                                                                                                Rest@7}}
                                                                                                    end
                                                                                                )
                                                                                        end,
                                                                                        fun(
                                                                                            _use0@8
                                                                                        ) ->
                                                                                            {Auth@1,
                                                                                                Rest@9} = _use0@8,
                                                                                            gleam@result:'try'(
                                                                                                begin
                                                                                                    _pipe@5 = Rest@9,
                                                                                                    must_equal(
                                                                                                        _pipe@5,
                                                                                                        <<>>
                                                                                                    )
                                                                                                end,
                                                                                                fun(
                                                                                                    _
                                                                                                ) ->
                                                                                                    Options = {connect_options,
                                                                                                        Clean_session
                                                                                                        =:= 1,
                                                                                                        Client_id,
                                                                                                        Keep_alive,
                                                                                                        Auth@1,
                                                                                                        Will@1},
                                                                                                    {ok,
                                                                                                        {Construct(
                                                                                                                Options
                                                                                                            ),
                                                                                                            Remainder}}
                                                                                                end
                                                                                            )
                                                                                        end
                                                                                    )
                                                                                end
                                                                            )
                                                                        end
                                                                    );

                                                                _ ->
                                                                    {error,
                                                                        invalid_data}
                                                            end
                                                        end
                                                    )
                                                end
                                            )
                                        end) end
                            )
                        end
                    )
                end
            )
        end
    ).

-file("src/spoke/internal/packet/decode.gleam", 152).
?DOC(false).
-spec publish(
    bitstring(),
    bitstring(),
    fun((spoke@internal@packet:publish_data()) -> OYV)
) -> {ok, {OYV, bitstring()}} | {error, decode_error()}.
publish(Flags, Data, Construct) ->
    case Flags of
        <<Dup:1, Qos:2, Retain:1>> ->
            gleam@result:'try'(
                decode_qos(Qos),
                fun(Qos@1) ->
                    gleam@result:'try'(
                        split_var_data(Data),
                        fun(_use0) ->
                            {Data@1, Remainder} = _use0,
                            gleam@result:'try'(
                                string(Data@1),
                                fun(_use0@1) ->
                                    {Topic, Data@2} = _use0@1,
                                    gleam@result:'try'(case Qos@1 of
                                            qo_s0 ->
                                                gleam@result:'try'(
                                                    begin
                                                        _pipe = Dup,
                                                        must_equal(_pipe, 0)
                                                    end,
                                                    fun(_) ->
                                                        {ok,
                                                            {fun(_capture) ->
                                                                    {publish_data_qo_s0,
                                                                        _capture}
                                                                end,
                                                                Data@2}}
                                                    end
                                                );

                                            qo_s1 ->
                                                gleam@result:'try'(
                                                    integer(Data@2),
                                                    fun(_use0@2) ->
                                                        {Id, Data@3} = _use0@2,
                                                        {ok,
                                                            {fun(_capture@1) ->
                                                                    {publish_data_qo_s1,
                                                                        _capture@1,
                                                                        Dup =:= 1,
                                                                        Id}
                                                                end,
                                                                Data@3}}
                                                    end
                                                );

                                            qo_s2 ->
                                                gleam@result:'try'(
                                                    integer(Data@2),
                                                    fun(_use0@3) ->
                                                        {Id@1, Data@4} = _use0@3,
                                                        {ok,
                                                            {fun(_capture@2) ->
                                                                    {publish_data_qo_s2,
                                                                        _capture@2,
                                                                        Dup =:= 1,
                                                                        Id@1}
                                                                end,
                                                                Data@4}}
                                                    end
                                                )
                                        end, fun(_use0@4) ->
                                            {Make_data, Data@5} = _use0@4,
                                            Msg_data = {message_data,
                                                Topic,
                                                Data@5,
                                                Retain =:= 1},
                                            {ok,
                                                {Construct(Make_data(Msg_data)),
                                                    Remainder}}
                                        end)
                                end
                            )
                        end
                    )
                end
            );

        _ ->
            {error, invalid_data}
    end.

-file("src/spoke/internal/packet/decode.gleam", 380).
?DOC(false).
-spec subscribe_request(bitstring()) -> {ok,
        {spoke@internal@packet:subscribe_request(), bitstring()}} |
    {error, decode_error()}.
subscribe_request(Bytes) ->
    gleam@result:'try'(
        string(Bytes),
        fun(_use0) ->
            {Topic_filter, Rest} = _use0,
            case Rest of
                <<Qos:8, Rest@1/binary>> ->
                    gleam@result:'try'(
                        decode_qos(Qos),
                        fun(Qos@1) ->
                            {ok,
                                {{subscribe_request, Topic_filter, Qos@1},
                                    Rest@1}}
                        end
                    );

                _ ->
                    {error, invalid_data}
            end
        end
    ).

-file("src/spoke/internal/packet/decode.gleam", 367).
?DOC(false).
-spec subscribe_requests(
    bitstring(),
    list(spoke@internal@packet:subscribe_request())
) -> {ok, list(spoke@internal@packet:subscribe_request())} |
    {error, decode_error()}.
subscribe_requests(Bytes, Requests) ->
    case Bytes of
        <<>> ->
            {ok, lists:reverse(Requests)};

        Bytes@1 ->
            gleam@result:'try'(
                subscribe_request(Bytes@1),
                fun(_use0) ->
                    {Request, Rest} = _use0,
                    _pipe = Rest,
                    subscribe_requests(_pipe, [Request | Requests])
                end
            )
    end.

-file("src/spoke/internal/packet/decode.gleam", 116).
?DOC(false).
-spec subscribe(
    bitstring(),
    bitstring(),
    fun((integer(), list(spoke@internal@packet:subscribe_request())) -> OYO)
) -> {ok, {OYO, bitstring()}} | {error, decode_error()}.
subscribe(Flags, Data, Construct) ->
    gleam@result:'try'(
        begin
            _pipe = Flags,
            must_equal(_pipe, <<2:4>>)
        end,
        fun(_) ->
            gleam@result:'try'(
                split_var_data(Data),
                fun(_use0) ->
                    {Data@1, Remainder} = _use0,
                    gleam@result:'try'(
                        integer(Data@1),
                        fun(_use0@1) ->
                            {Packet_id, Rest} = _use0@1,
                            gleam@result:'try'(
                                subscribe_requests(Rest, []),
                                fun(Requests) ->
                                    {ok,
                                        {Construct(Packet_id, Requests),
                                            Remainder}}
                                end
                            )
                        end
                    )
                end
            )
        end
    ).
