-module(spoke@internal@packet@encode).
-compile([no_auto_import, nowarn_unused_vars, nowarn_unused_function, nowarn_nomatch]).

-export([string/1, varint/1, unsubscribe/2, only_packet_type/1, only_packet_id/3, subscribe/2, suback/2, connack/1, publish/1, connect/1]).
-export_type([encode_error/0]).

-if(?OTP_RELEASE >= 27).
-define(MODULEDOC(Str), -moduledoc(Str)).
-define(DOC(Str), -doc(Str)).
-else.
-define(MODULEDOC(Str), -compile([])).
-define(DOC(Str), -compile([])).
-endif.

?MODULEDOC(false).

-type encode_error() :: empty_subscribe_list | empty_unsubscribe_list.

-file("src/spoke/internal/packet/encode.gleam", 179).
?DOC(false).
-spec string(binary()) -> bitstring().
string(Str) ->
    Data = gleam_stdlib:identity(Str),
    Len = <<(erlang:byte_size(Data)):16/big>>,
    gleam_stdlib:bit_array_concat([Len, Data]).

-file("src/spoke/internal/packet/encode.gleam", 190).
?DOC(false).
-spec build_varint(bitstring(), integer()) -> bitstring().
build_varint(Bits, I) ->
    case I < 128 of
        true ->
            <<Bits/bitstring, I:8>>;

        false ->
            Remainder = I rem 128,
            _pipe = <<Bits/bitstring, 1:1, Remainder:7>>,
            build_varint(_pipe, I div 128)
    end.

-file("src/spoke/internal/packet/encode.gleam", 185).
?DOC(false).
-spec varint(integer()) -> bitstring().
varint(I) ->
    _pipe = <<>>,
    build_varint(_pipe, I).

-file("src/spoke/internal/packet/encode.gleam", 200).
?DOC(false).
-spec encode_parts(
    integer(),
    bitstring(),
    bitstring(),
    gleam@bytes_tree:bytes_tree()
) -> gleam@bytes_tree:bytes_tree().
encode_parts(Id, Flags, Header, Payload) ->
    Variable_content = begin
        _pipe = Payload,
        gleam@bytes_tree:prepend(_pipe, Header)
    end,
    Remaining_len = erlang:iolist_size(Variable_content),
    Fixed_header = <<Id:4, Flags/bitstring, (varint(Remaining_len))/bitstring>>,
    gleam@bytes_tree:prepend(Variable_content, Fixed_header).

-file("src/spoke/internal/packet/encode.gleam", 138).
?DOC(false).
-spec unsubscribe(integer(), list(binary())) -> {ok,
        gleam@bytes_tree:bytes_tree()} |
    {error, encode_error()}.
unsubscribe(Packet_id, Topics) ->
    case Topics of
        [] ->
            {error, empty_unsubscribe_list};

        _ ->
            Header = <<Packet_id:16/big>>,
            Payload = begin
                gleam@list:fold(
                    Topics,
                    gleam@bytes_tree:new(),
                    fun(Builder, Topic) -> _pipe = Builder,
                        gleam@bytes_tree:append(_pipe, string(Topic)) end
                )
            end,
            {ok, encode_parts(10, <<2:4>>, Header, Payload)}
    end.

-file("src/spoke/internal/packet/encode.gleam", 166).
?DOC(false).
-spec only_packet_type(integer()) -> gleam@bytes_tree:bytes_tree().
only_packet_type(Type_id) ->
    encode_parts(Type_id, <<0:4>>, <<>>, gleam@bytes_tree:new()).

-file("src/spoke/internal/packet/encode.gleam", 170).
?DOC(false).
-spec only_packet_id(integer(), integer(), integer()) -> gleam@bytes_tree:bytes_tree().
only_packet_id(Type_id, Flags, Packet_id) ->
    encode_parts(
        Type_id,
        <<Flags:4>>,
        <<Packet_id:16/big>>,
        gleam@bytes_tree:new()
    ).

-file("src/spoke/internal/packet/encode.gleam", 214).
?DOC(false).
-spec encode_qos(spoke@internal@packet:qo_s()) -> integer().
encode_qos(Qos) ->
    case Qos of
        qo_s0 ->
            0;

        qo_s1 ->
            1;

        qo_s2 ->
            2
    end.

-file("src/spoke/internal/packet/encode.gleam", 99).
?DOC(false).
-spec subscribe(integer(), list(spoke@internal@packet:subscribe_request())) -> {ok,
        gleam@bytes_tree:bytes_tree()} |
    {error, encode_error()}.
subscribe(Packet_id, Topics) ->
    case Topics of
        [] ->
            {error, empty_subscribe_list};

        _ ->
            Header = <<Packet_id:16/big>>,
            Payload = begin
                gleam@list:fold(
                    Topics,
                    gleam@bytes_tree:new(),
                    fun(Builder, Topic) -> _pipe = Builder,
                        _pipe@1 = gleam@bytes_tree:append(
                            _pipe,
                            string(erlang:element(2, Topic))
                        ),
                        gleam@bytes_tree:append(
                            _pipe@1,
                            <<(encode_qos(erlang:element(3, Topic))):8>>
                        ) end
                )
            end,
            {ok, encode_parts(8, <<2:4>>, Header, Payload)}
    end.

-file("src/spoke/internal/packet/encode.gleam", 222).
?DOC(false).
-spec encode_suback_result({ok, spoke@internal@packet:qo_s()} | {error, nil}) -> integer().
encode_suback_result(Result) ->
    case Result of
        {ok, Qos} ->
            encode_qos(Qos);

        {error, _} ->
            16#80
    end.

-file("src/spoke/internal/packet/encode.gleam", 119).
?DOC(false).
-spec suback(integer(), list({ok, spoke@internal@packet:qo_s()} | {error, nil})) -> {ok,
        gleam@bytes_tree:bytes_tree()} |
    {error, encode_error()}.
suback(Packet_id, Results) ->
    case Results of
        [] ->
            {error, empty_subscribe_list};

        _ ->
            Header = <<Packet_id:16/big>>,
            Payload = begin
                gleam@list:fold(
                    Results,
                    gleam@bytes_tree:new(),
                    fun(Builder, Result) -> _pipe = Builder,
                        gleam@bytes_tree:append(
                            _pipe,
                            <<(encode_suback_result(Result)):8>>
                        ) end
                )
            end,
            {ok, encode_parts(9, <<0:4>>, Header, Payload)}
    end.

-file("src/spoke/internal/packet/encode.gleam", 229).
?DOC(false).
-spec encode_connack_error(spoke@internal@packet:connect_error()) -> integer().
encode_connack_error(Code) ->
    case Code of
        unacceptable_protocol_version ->
            1;

        identifier_refused ->
            2;

        server_unavailable ->
            3;

        bad_username_or_password ->
            4;

        not_authorized ->
            5
    end.

-file("src/spoke/internal/packet/encode.gleam", 157).
?DOC(false).
-spec connack({ok, boolean()} | {error, spoke@internal@packet:connect_error()}) -> gleam@bytes_tree:bytes_tree().
connack(Result) ->
    Header = case Result of
        {ok, true} ->
            <<1, 0>>;

        {ok, false} ->
            <<0, 0>>;

        {error, E} ->
            <<0, (encode_connack_error(E)):8>>
    end,
    encode_parts(2, <<0:4>>, Header, gleam@bytes_tree:new()).

-file("src/spoke/internal/packet/encode.gleam", 239).
?DOC(false).
-spec option_to_flag(gleam@option:option(any())) -> integer().
option_to_flag(Option) ->
    case Option of
        {some, _} ->
            1;

        none ->
            0
    end.

-file("src/spoke/internal/packet/encode.gleam", 246).
?DOC(false).
-spec encode_bool(boolean()) -> integer().
encode_bool(Val) ->
    case Val of
        true ->
            1;

        false ->
            0
    end.

-file("src/spoke/internal/packet/encode.gleam", 253).
?DOC(false).
-spec integer(integer()) -> bitstring().
integer(Val) ->
    <<Val:16/big>>.

-file("src/spoke/internal/packet/encode.gleam", 83).
?DOC(false).
-spec publish(spoke@internal@packet:publish_data()) -> gleam@bytes_tree:bytes_tree().
publish(Data) ->
    {Msg@3, Dup@2, Packet_id, Qos} = case Data of
        {publish_data_qo_s0, Msg} ->
            {Msg, false, <<>>, qo_s0};

        {publish_data_qo_s1, Msg@1, Dup, Id} ->
            {Msg@1, Dup, integer(Id), qo_s1};

        {publish_data_qo_s2, Msg@2, Dup@1, Id@1} ->
            {Msg@2, Dup@1, integer(Id@1), qo_s2}
    end,
    Dup@3 = encode_bool(Dup@2),
    Retain = encode_bool(erlang:element(4, Msg@3)),
    Flags = <<Dup@3:1, (encode_qos(Qos)):2, Retain:1>>,
    Header = <<(string(erlang:element(2, Msg@3)))/bitstring,
        Packet_id/bitstring>>,
    Payload = gleam@bytes_tree:from_bit_array(erlang:element(3, Msg@3)),
    encode_parts(3, Flags, Header, Payload).

-file("src/spoke/internal/packet/encode.gleam", 257).
?DOC(false).
-spec encode_or_zero(gleam@option:option(PSU), fun((PSU) -> integer())) -> integer().
encode_or_zero(Option, Map) ->
    case Option of
        {some, X} ->
            Map(X);

        none ->
            0
    end.

-file("src/spoke/internal/packet/encode.gleam", 19).
?DOC(false).
-spec connect(spoke@internal@packet:connect_options()) -> gleam@bytes_tree:bytes_tree().
connect(Options) ->
    Password = gleam@option:then(
        erlang:element(5, Options),
        fun(A) -> erlang:element(3, A) end
    ),
    Will_retain = gleam@option:map(
        erlang:element(6, Options),
        fun(W) -> erlang:element(4, (erlang:element(1, W))) end
    ),
    Will_qos = gleam@option:map(
        erlang:element(6, Options),
        fun(W@1) -> erlang:element(2, W@1) end
    ),
    Header = <<(string(<<"MQTT"/utf8>>))/bitstring,
        4:8,
        (option_to_flag(erlang:element(5, Options))):1,
        (option_to_flag(Password)):1,
        (encode_or_zero(Will_retain, fun encode_bool/1)):1,
        (encode_or_zero(Will_qos, fun encode_qos/1)):2,
        (option_to_flag(erlang:element(6, Options))):1,
        (encode_bool(erlang:element(2, Options))):1,
        0:1,
        (erlang:element(4, Options)):16/big>>,
    Payload = begin
        _pipe = gleam@bytes_tree:new(),
        gleam@bytes_tree:append(_pipe, string(erlang:element(3, Options)))
    end,
    Payload@1 = case erlang:element(6, Options) of
        {some, {Will, _}} ->
            Size = erlang:byte_size(erlang:element(3, Will)),
            _pipe@1 = Payload,
            _pipe@2 = gleam@bytes_tree:append(
                _pipe@1,
                string(erlang:element(2, Will))
            ),
            _pipe@3 = gleam@bytes_tree:append(_pipe@2, <<Size:16/big>>),
            gleam@bytes_tree:append(_pipe@3, erlang:element(3, Will));

        none ->
            Payload
    end,
    Payload@3 = case erlang:element(5, Options) of
        {some, Auth} ->
            Payload@2 = begin
                _pipe@4 = Payload@1,
                gleam@bytes_tree:append(
                    _pipe@4,
                    string(erlang:element(2, Auth))
                )
            end,
            case erlang:element(3, Auth) of
                {some, Password@1} ->
                    Len = erlang:byte_size(Password@1),
                    _pipe@5 = Payload@2,
                    _pipe@6 = gleam@bytes_tree:append(_pipe@5, <<Len:16/big>>),
                    gleam@bytes_tree:append(_pipe@6, Password@1);

                none ->
                    Payload@2
            end;

        none ->
            Payload@1
    end,
    encode_parts(1, <<0:4>>, Header, Payload@3).
